"""
Dynamic visualization viewing modality module
- Tools for creating different viewing modalities for the dynamic visualizations
"""

# author: Eric David Sun <edsun@stanford.edu>
# (C) 2022 
from __future__ import print_function, division

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from PIL import Image
import cv2
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
import plotly.express as px



def interactive(df, label, show=False, save=False, colors=None,
            xlabel="x1", ylabel="x2", title="", legend_title="",
            vmin=None, vmax=None, alpha=1.0, width=4, height=4, dpi=500, dim=None):
    '''
    Generate dynamic visualization with interactive (plotly) modality
    
    Arguments:
        df = pandas dataframe corresponding to output of boot.generate() 
        label = str or float, column in output for labeling/coloring visualization
        show = True or False, whether to show the visualization
        save = False or str, path to save the visualization (acceptable extensions are .html)
        colors = list of plotly color codes to use if labels are discrete
        xlabel = label of x-axis
        ylabel = label of y-axis
        title = title of plot
        legend_title = title of legend
        vmin = set value of vmin for continuous colorbar (default is min across all bootstrap frames)
        vmax = set value of vmax for continuous colorbar (default is max across all bootstrap frames)
        alpha = opacity between 0 and 1
        width = width of plotly plot
        height = height of plotly plot
        dpi = dots per inch
        dim = x and y dimensions will +- dim in either direction; default is to take the maximum across all frames
        
        will convert width, height, dpi into plotly.express.scatter() width and height
        
    Returns:
        fig = Plotly object
    '''
    # run some checks
    if isinstance(colors, list):
        if len(colors) < len(np.unique(df[label])):
            raise Exception ("colors need to have >= length as len(np.unique(df[label]))")
    
    # get bootstrap-wide statistics for consistency across frames
    if dim is None:
        dim = np.max([np.max(np.abs(df['x1'])),np.max(np.abs(df['x2']))])
    
    
    if df[label].dtype == np.float: # define max and min values across all frames
        if vmin is None:
            vmin = np.min(df[label])
        if vmax is None:
            vmax = np.max(df[label])
            
    # compute width and height
    width = width * dpi
    height = height * dpi
    
                
    df_px = df.copy() # make copy of df
    
    # create color dict if needed
    if (df_px[label].dtype != np.float) and (isinstance(colors, list)):
        color_dict = {}
        for li, lab in enumerate(np.unique(df_px[label])):
            color_dict[lab] = colors[li]
        
    # add new columns for marker
    #df_px['symbol'] = marker
    
    # make figure
    if df[label].dtype == np.float: # continuous labels
        fig = px.scatter(df_px, x='x1', y='x2', animation_frame="bootstrap_number", animation_group="original_index", color=label, hover_name=label, title=title,
                   range_x=[-dim,dim], range_y=[-dim,dim], range_color=(vmin,vmax), opacity=alpha, width=width, height=height,
                   labels={
                         "x1": xlabel,
                         "x2": ylabel,
                         label: legend_title
                     })
    elif isinstance(colors, list): # if colors is specified for discrete labels           
        fig = px.scatter(df_px, x='x1', y='x2', animation_frame="bootstrap_number", animation_group="original_index", color=label, hover_name=label, title=title,
                   range_x=[-dim,dim], range_y=[-dim,dim], opacity=alpha, width=width, height=height,
                   color_discrete_map = color_dict,
                   labels={
                         "x1": xlabel,
                         "x2": ylabel,
                         label: legend_title
                     })
                     
    else: # discrete labels (default colors)
        fig = px.scatter(df_px, x='x1', y='x2', animation_frame="bootstrap_number", animation_group="original_index", color=label, hover_name=label, title=title,
                   range_x=[-dim,dim], range_y=[-dim,dim], opacity=alpha, width=width, height=height,
                   labels={
                         "x1": xlabel,
                         "x2": ylabel,
                         label: legend_title
                     })

    if isinstance(save, str): # save as HTML
        if ".html" in save:
            fig.write_html(save)
        else:
            raise Exception("save needs to be .html file")
    
    if show is True: # show figure
        fig.show()
    
    return(fig)
    
    
def animated(df, label, save=False, get_frames=None, colors=None, cmap='viridis', width=4, height=4, dpi=500,
            xlabel=None, ylabel=None, title=None, title_fontsize=16, major_fontsize=16, minor_fontsize=14,
            vmin=None, vmax=None, marker="o", alpha=1.0, solid_cbar=True, show_legend=True, solid_legend=True,
            legend_fontsize=12, dim=None, **kwargs):
    '''
    Generate dynamic visualization with animated modality
        
    Arguments:
        df = pandas dataframe corresponding to output of boot.generate() 
        label = str, column in output for labeling/coloring visualization
        save = False or str, path to save the visualization (acceptable extensions are .gif and .avi)
        get_frames = None or list of integer indices to save frames for in folder called save.split(".")[0]
        colors = list of matplotlib color codes to use if labels are discrete
        cmap = color map to use for continuous labels
        width = width of matplotlib plot
        height = height of matplotlib plot
        dpi = dots per inch
        xlabel = label of x-axis
        ylabel = label of y-axis
        title = title of plot
        title_fontsize = fontsize for title
        major_fontsize = fontsize for axis labels
        minor_fontsize = fontsize for axis ticks
        vmin = set value of vmin for continuous colorbar (default is min across all bootstrap frames)
        vmax = set value of vmax for continuous colorbar (default is max across all bootstrap frames)
        marker = marker symbol for plot
        alpha = opacity between 0 and 1
        solid_cbar = boolean, whether to set alpha=1 for color bar
        show_legend = boolean, whether to show a legend
        solid_legend = boolean, whether to set alpha=1 for legend markers
        legend_fontsize = fontsize for legend
        dim = x and y dimensions will +- dim in either direction; default is to take the maximum across all frames
                
    Returns:
        images = list of matplotlib objects
    '''
    # run some checks
    if isinstance(colors, list):
        if len(colors) < len(np.unique(df[label])):
            raise Exception ("colors need to have >= length as len(np.unique(df[label]))")
    
    # get bootstrap-wide statistics for consistency across frames
    if dim is None:
        dim = np.max([np.max(np.abs(df['x1'])),np.max(np.abs(df['x2']))])
    if df[label].dtype == np.float:
        if vmin is None:
            vmin = np.min(df[label])
        if vmax is None:
            vmax = np.max(df[label])
    
    # generate list of matplotlib plots
    images = []
    for boot_num in np.sort(np.unique(df["bootstrap_number"])):
    
        # get bootstrap results
        sub_df = df[df["bootstrap_number"] == boot_num].reset_index()
        
        # make figure
        fig = plt.figure(figsize=(width, height))
        
        # continuous labels for float labels
        if df[label].dtype == np.float:
            plt.scatter(sub_df["x1"], sub_df["x2"], c=sub_df[label], cmap=cmap, vmin=vmin, vmax=vmax,
                        marker=marker, alpha=alpha, **kwargs)
            cbar = plt.colorbar()
            if solid_cbar is True:
                cbar.solids.set(alpha=1)
            
        # otherwise, discrete labels
        else:
            for li, lab in enumerate(np.unique(df[label])):
                if colors is None:
                    plt.scatter(sub_df[sub_df[label] == lab]["x1"], sub_df[sub_df[label] == lab]["x2"],
                                label=str(lab), marker=marker, alpha=alpha, **kwargs)
                else:
                    plt.scatter(sub_df[sub_df[label] == lab]["x1"], sub_df[sub_df[label] == lab]["x2"],
                                label=str(lab), c=colors[li], marker=marker, alpha=alpha, **kwargs)
                                
            if show_legend is True:
                leg = plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=legend_fontsize)
                if solid_legend is True:
                    for lh in leg.legendHandles: 
                        lh.set_alpha(1)
        
        # add titles and axis labels
        if isinstance(title, str):
            plt.title(title, fontsize=title_fontsize)
        if isinstance(xlabel, str):
            plt.xlabel(xlabel, fontsize=major_fontsize)
        if isinstance(ylabel, str):
            plt.ylabel(ylabel, fontsize=major_fontsize)
            
        plt.xticks(fontsize=minor_fontsize)
        plt.yticks(fontsize=minor_fontsize)
            
        # specify plot parameters
        plt.ylim(-dim,dim)
        plt.xlim(-dim,dim)
        plt.tight_layout()
        
        # save still frames if specified
        if isinstance(get_frames, list):
            if not os.path.exists(save.split(".")[0]):
                os.makedirs(save.split(".")[0])
            #for index in get_frames:
            #    im = images[index]
            #    im.save(os.path.join(save.split(".")[0],"frame"+str(index)+".pdf"))
            if boot_num in get_frames:
                plt.savefig(os.path.join(save.split(".")[0],"frame"+str(boot_num)+".pdf"), dpi=dpi)

        # make plot into image
        im = fig2img(fig, dpi=dpi)
        images.append(im)
        plt.close()
    
    # save dynamic visualization
    if isinstance(save, str):
        # save as GIF
        if ".gif" in save:
            images[0].save(save, save_all = True, append_images = images[1:], 
                           optimize = False, duration = [1000]+[100]*len(images[1:]), loop=0)
        # save as AVI
        elif ".avi" in save:
            frame = cv2.cvtColor(np.array(images[0]), cv2.COLOR_RGB2BGR)
            height, width, layers = frame.shape
            video = cv2.VideoWriter(save, 0, 10, (width,height))
            for image in images:
                opencvImage = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
                video.write(opencvImage)
            cv2.destroyAllWindows()
            video.release()
        else:
            raise Exception("save needs to be .gif or .avi file")
    
    return(images)
    
    
def stacked(df, label, show=False, save=False, colors=None, cmap='viridis', width=4, height=4, dpi=500,
            xlabel=None, ylabel=None, title=None, title_fontsize=16, major_fontsize=16, minor_fontsize=14,
            vmin=None, vmax=None, marker="o", alpha=None, solid_cbar=True, show_legend=True, solid_legend=True,
            legend_fontsize=12, dim=None, frame=None, return_fig=True, **kwargs):
    '''
    Generate dynamic visualization with stacked modality
        
    Arguments:
        df = pandas dataframe corresponding to output of boot.generate() 
        label = str, column in output for labeling/coloring visualization
        show = True or False, whether to show the visualization
        save = False or str, path to save the visualization (acceptable extensions are .png, .jpg, .pdf, .eps, .tiff, ...
        frame = None or if int, then shows only that bootstrap frame
        return_fig = boolean, whether to return figure or not
        
        See animated() for more details
                
    Returns:
        fig = matplotlib object
    '''
    # run some checks
    if isinstance(colors, list):
        if len(colors) < len(np.unique(df[label])):
            raise Exception ("colors need to have >= length as len(np.unique(df[label]))")
    
    # get bootstrap-wide statistics for consistency across frames
    if dim is None:
        dim = np.max([np.max(np.abs(df['x1'])),np.max(np.abs(df['x2']))])
    if df[label].dtype == np.float:
        if vmin is None:
            vmin = np.min(df[label])
        if vmax is None:
            vmax = np.max(df[label])
        
    # automatically set alpha based on heuristic
    if alpha is None:
        alpha = 0.2/(df.shape[0]/1000)

    fig = plt.figure(figsize=(width,height))
    
    # snapshot frame if requested
    if isinstance(frame, int):
        df = df[df['bootstrap_number']==frame]
    elif isinstance(frame, float):
        df = df[df['bootstrap_number']==int(frame)]
    
    # continuous labels
    if df[label].dtype == np.float:
        plt.scatter(df["x1"],df["x2"], c=df[label], cmap=cmap, vmin=vmin, vmax=vmax,
                        marker=marker, alpha=alpha, edgecolors='none', **kwargs)
        cbar = plt.colorbar()
        if solid_cbar is True:
            cbar.solids.set(alpha=1)
    
    # discrete labels
    else:
        for li, lab in enumerate(np.unique(df[label])):
            if colors is None:
                plt.scatter(df[df[label] == lab]["x1"],df[df[label] == lab]["x2"],
                            label=str(lab), marker=marker, alpha=alpha, edgecolors='none', **kwargs)
            else:
                plt.scatter(df[df[label] == lab]["x1"],df[df[label] == lab]["x2"],
                            label=str(lab), c=colors[li], marker=marker, alpha=alpha, edgecolors='none', **kwargs)

        if show_legend is True:
            leg = plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=legend_fontsize)
            if solid_legend is True:
                for lh in leg.legendHandles: 
                    lh.set_alpha(1)
                    
    # add titles and axis labels
    if isinstance(title, str):
        plt.title(title, fontsize=title_fontsize)
    if isinstance(xlabel, str):
        plt.xlabel(xlabel, fontsize=major_fontsize)
    if isinstance(ylabel, str):
        plt.ylabel(ylabel, fontsize=major_fontsize)
        
    plt.xticks(fontsize=minor_fontsize)
    plt.yticks(fontsize=minor_fontsize)
                    
    plt.ylim(-dim,dim)
    plt.xlim(-dim,dim)
    plt.tight_layout()
    
    # save plot
    if isinstance(save, str):
        plt.savefig(save, dpi=dpi, bbox_inches = "tight")
    if show is True:
        plt.show()
    else:
        plt.close()
    
    if return_fig is True:
        return(fig)
    


def fig2img(fig, dpi):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    import io
    buf = io.BytesIO()
    fig.savefig(buf, dpi=dpi, bbox_inches="tight")
    buf.seek(0)
    img = Image.open(buf)
    return img
    
    
    



"""
Score module
- Tools for computing variance score and stability score
- Tools for computing concordance score and ensemble concordance score
"""

# author: Eric David Sun <edsun@stanford.edu>
# (C) 2022 
from __future__ import print_function, division

from sklearn.decomposition import TruncatedSVD
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import NearestNeighbors
from scipy.linalg import eigh
from scipy.stats import pearsonr, spearmanr
import numpy as np
import os
#from numba import njit, jit, prange
from tqdm import tqdm
from joblib import Parallel, delayed
import warnings
warnings.filterwarnings("ignore")
import time



# Variance Score
def variance (df, method="global", k=20, X_orig=None, neighborhoods=None, normalize_pairwise_distance=False,
                include_original=True, return_times=False):
    '''
    Computes variances scores from the output dataframe (out) of boot.generate()
    
    Arguments:
        df = pandas dataframe, output of boot.generate()
        method = str, specifies the type of stability score to compute
            "global" - compute stability score across the entire dataset
            "random" - approximate global stability score by randomly selecting k "neighbors" for each observation
            "local" - compute stability over k-nearest neighbors (specify k, defaults to 20)
        k = int, when method is "local" or "random"; specifies the number of neighbors
        neighborhoods = array or list of size n, with elements labeling neighborhoods
        normalize_pairwise_distance = if True, then divide each set of d[i,j] by its mean before computing variance
        include_original = if True, include the original (bootstrap_number=-1) embedding in calculating scores
        return_times = True or False; if not False, returns a dictionary of run times broken down by components as the second output
        
    Returns:
        mean_variance_distances = numpy array with variance scores (mean variance in pairwise distance to neighborhood) for each observation
    '''
    # keep track of run times
    rt_dict = {}
    
    # retrieve embeddings and bootstrap indices
    if include_original is True:
        embeddings = [np.array(df[df["bootstrap_number"]==b][["x1","x2"]].values) for b in np.unique(df["bootstrap_number"])]
        bootstrap_indices = [np.array(df[df["bootstrap_number"]==b]["original_index"].values) for b in np.unique(df["bootstrap_number"])]
    else:
        embeddings = [np.array(df[df["bootstrap_number"]==b][["x1","x2"]].values) for b in np.unique(df["bootstrap_number"]) if b!=-1]
        bootstrap_indices = [np.array(df[df["bootstrap_number"]==b]["original_index"].values) for b in np.unique(df["bootstrap_number"]) if b!=-1]
    
    # set up neighborhoods for variance score
    print("Setting up neighborhoods...")
    start_time = time.time()
    neighborhood_dict = get_neighborhood_dict(method, k, keys=np.unique(df["original_index"]), neighborhoods=neighborhoods, X_orig=X_orig)
    neighborhood_time = time.time() - start_time
    rt_dict["neighborhood"] = neighborhood_time
    print("--- %s seconds ---" % neighborhood_time)
    
    # populate distance dict
    print("Populating distances...")
    start_time = time.time()
    dist_dict = populate_distance_dict(neighborhood_dict, embeddings, bootstrap_indices)
    dist_time = time.time() - start_time
    rt_dict["distances"] = dist_time
    print("--- %s seconds ---" % dist_time)
    
    # compute mean pairwise distance for normalization
    print("Computing mean pairwise distance for normalization...")
    start_time = time.time()
    mean_pairwise_distance = compute_mean_distance(dist_dict, normalize_pairwise_distance)
    norm_time = time.time() - start_time
    rt_dict["normalization"] = norm_time
    print("--- %s seconds ---" % norm_time)

    # compute variances
    print("Computing variance scores...")
    start_time = time.time()
    mean_variance_distances = compute_mean_variance_distance(dist_dict, normalize_pairwise_distance, mean_pairwise_distance)
    var_time = time.time() - start_time
    rt_dict["variance"] = var_time
    print("--- %s seconds ---" % var_time)
    
    if return_times is False:
        return(mean_variance_distances)
    else:
        return(mean_variance_distances, rt_dict)


def stability_from_variance(mean_variance_distances, alpha):
    '''
    For alpha and mean_variance_distances, computes stability scores
    
    Arguments:
        mean_variance_distances = list or array of variance scores (output of variance())
        alpha = float > 0, is the exponential paramter for stability score formula: stability = 1/(1+variance)^alpha
            defaults to alpha=1.0
        
    Returns:
        stability_scores = numpy array with stability score for each observation
    '''
    # compute stability score
    print("Computing stability score with alpha="+str(alpha)+" ...")
    start_time = time.time()
    stability_scores = 1 / (1 + mean_variance_distances)**alpha
    print("--- %s seconds ---" % (time.time() - start_time))
    
    return(stability_scores)


# Stability score
def stability (df, method="global", alpha=1.0, k=20, X_orig=None, neighborhoods=None, normalize_pairwise_distance=False,
                include_original=True, return_times=False):
    '''
    Computes stability scores from the output dataframe (out) of boot.generate()
    
    Arguments:
        alpha = float > 0, is the exponential paramter for stability score formula: stability = 1/(1+variance)^alpha
            defaults to alpha=1.0
        See variance() for more details
        
    Returns:
        stability_scores = numpy array with stability score for each observation
    '''
    # check alpha > 0
    if alpha <= 0:
        raise Exception("alpha must be >= 0")
    
    if return_times is False:
        mean_variance_distances = variance(df, method=method, k=k, X_orig=X_orig, neighborhoods=neighborhoods,
                                            normalize_pairwise_distance=normalize_pairwise_distance,
                                            include_original=include_original, return_times=return_times)
    else:
        mean_variance_distances, rt_dict = variance(df, method=method, k=k, X_orig=X_orig, neighborhoods=neighborhoods,
                                            normalize_pairwise_distance=normalize_pairwise_distance,
                                            include_original=include_original, return_times=return_times)

    # compute stability score
    stability_scores = stability_from_variance(mean_variance_distances, alpha)
    
    if return_times is False:
        return(stability_scores)
    else:
        return(stability_scores, rt_dict)




#@njit(parallel=True)
def get_neighborhood_dict(method, k, keys, neighborhoods=None, X_orig=None):
    '''
    Returns a neighborhood dictionary where keys are observation indices
    
    Arguments:
        method = string specifier for the method for constructing neighborhoods ("global", "random", "local")
            to specify a predefined/custom neighborhood, use the neighborhoods argument
        k = integer specfiying size of neighborhood for "local" and "random" methods
        keys = list/array of the key names (integers) to use in constructing the dictionary
        neighborhoods = predefined neighborhoods [list of strings where same strings indicate observations in the same neighborhood]
        X_orig = original X used to computing the bootstraps; used to determine local kNN relations
        
    Returns:
        neighborhood_dict = dictionary with keys corresponding to observation indices and lists of their neighborhoods as values
    '''
    neighborhood_dict = dict()
    for key in keys:
        neighborhood_dict[str(key)] = []
        
    if neighborhoods is None:
    
        if method == "global":
            for n in range(len(neighborhood_dict.keys())):
                key = list(neighborhood_dict.keys())[n]
                neighborhood_dict[key] = [i for i in range(len(neighborhood_dict.keys()))]
        
        elif method == "random":
            for n in range(len(neighborhood_dict.keys())):
                key = list(neighborhood_dict.keys())[n]
                neighborhood_dict[key] = [i for i in np.random.randint(0,len(neighborhood_dict.keys()),k)]
        
        elif method == "local":
            if X_orig is None:
                raise Exception("Need to specify X_orig to compute nearest neighbors")
            for n in range(len(neighborhood_dict.keys())):
                key = list(neighborhood_dict.keys())[n]
                nbrs = NearestNeighbors(n_neighbors=k+1).fit(X_orig) # compute w.r.t. reference
                distances_local, indices_local = nbrs.kneighbors(X_orig)
                neighborhood_dict[key] = [i for i in indices_local[int(key),1:k+1]]
        
        else:
            raise Exception ("Need to specify either method ('global', 'local') or neighborhood")  
    
    else:
        for n in range(len(neighborhood_dict.keys())):
            key = list(neighborhood_dict.keys())[n]
            label = neighborhoods[int(key)]
            neighborhood_dict[key] = [i for i in range(len(neighborhoods)) if neighborhoods[i]==label]
            
            
    return(neighborhood_dict)

#@njit(parallel=True)
def populate_distance_dict (neighborhood_dict, embeddings, bootstrap_indices):
    '''
    Returns dictionary with pairwise dictionaries for all observations[i][j]
    
    Arguments:
        neighborhood_dict = dictionary with keys corresponding to observation indices and lists of their neighborhoods as values
            this is the output of get_neighborhood_dict()
        embeddings = list of nx2 DR visualization arrays [see outputs of boot.generate()]
        bootstrap_indices = list of arrays, where each array is the bootstrap indices of that visualization [see outputs of boot.generate()]
        
    Returns:
        dist_dict = two-level dictionary with first level corresponding to observation i and second level corresponding to observation j 
                    the value corresponds to a list of Euclidean distances between observation i and j across all co-occurrences in the bootstrap visualizations
    '''
    dist_dict = dict()
    for key1 in neighborhood_dict.keys():
        dist_dict[str(key1)] = {}
        for key2 in neighborhood_dict[key1]:
            dist_dict[str(key1)][str(key2)] = []
            
    for b in range(len(embeddings)):
        dist_mat = pairwise_distances(embeddings[b], n_jobs=-1)
        boot_idxs = bootstrap_indices[b]
        
        for i in range(dist_mat.shape[0]):
            key1 = str(boot_idxs[i])
            neighbor_js = neighborhood_dict[key1]

            for nj in neighbor_js:
                key2 = str(nj)
                js = [x[0] for x in np.argwhere(boot_idxs == nj)]
                for j in js:
                    dist_dict[key1][key2].append(dist_mat[i,j])

    return(dist_dict)

#@njit(parallel=True)
def compute_mean_distance(dist_dict, normalize_pairwise_distance=False):
    '''
    Computes mean pairwise distance across all (i,j)
    
    Arguments:
        dist_dict = output of populate_distance_dict()
        normalize_pairwise_distance = boolean; whether to normalize the distances between i and j by their mean
    
    Returns:
        mean_pairwise_distance = the mean of all distances between any i and any j; used to normalize/scale the variance score
    '''
    all_distances = []
    
    for n in range(len(dist_dict.keys())):
        key1 = list(dist_dict.keys())[n]
        
        for key2 in dist_dict[key1].keys():
            distances = np.array(dist_dict[key1][key2])
            
            if normalize_pairwise_distance is True: # perform additional local normalization before taking variance
                distances = distances/(np.nanmean(distances)) # recall: V(CX) = C^2 V(X)
            all_distances.append(np.nanmean(distances))
            
    mean_pairwise_distance = np.nanmean(all_distances)
    
    return(mean_pairwise_distance)

#@njit(parallel=True)
def compute_mean_variance_distance(dist_dict, normalize_pairwise_distance=False, mean_pairwise_distance=1.0):
    '''
    For each (i,j) compute the variance across all distances.
    Then for each i, average across all var(i,j)
    
    Arguments:
        dist_dict = output of populate_distance_dict()
        normalize_pairwise_distance = boolean; whether to normalize the distances between i and j by their mean
        mean_pairwise_distance = float, output of compute_mean_distance()
    
    Returns:
        mean_variance_distances = list of variance scores [one for each observation]
    '''
    mean_variance_distances = np.ones(len(dist_dict.keys()))*np.inf
    
    for n in range(len(dist_dict.keys())):
        key1 = list(dist_dict.keys())[n]
        variances = []
        
        for key2 in dist_dict[key1].keys():
            distances = np.array(dist_dict[key1][key2])
            
            if normalize_pairwise_distance is True: # perform additional local normalization before taking variance
                distances = distances/(np.nanmean(distances))
                
            variances.append(np.nanvar(distances / mean_pairwise_distance)) # normalize globally and compute variance
        
        mean_variance_distances[int(key1)] = np.nanmean(variances)
        
    return(mean_variance_distances)


        
### CONCORDANCE SCORES -- see our publication for mathematical details

def get_jaccard(X_orig, X_red, k, precomputed=[False, False]):
    '''
    Computes Jaccard coefficient at k for each point in X_orig and X_red
    '''
    # INIT NEAREST NEIGHBORS
    if precomputed[0] is False:
        nbrs = NearestNeighbors(n_neighbors=k).fit(X_orig)
    else:
        nbrs = NearestNeighbors(n_neighbors=k, metric='precomputed').fit(X_orig)
    distances_orig, indices_orig = nbrs.kneighbors(X_orig)
    
    if precomputed[0] is False:
        nbrs = NearestNeighbors(n_neighbors=k).fit(X_red)
    else:
        nbrs = NearestNeighbors(n_neighbors=k, metric='precomputed').fit(X_red)
    distances_red, indices_red = nbrs.kneighbors(X_red)
    
    # COMPUTE JACCARD
    jaccards = []
    for i in range(X_orig.shape[0]):
        list1 = list(indices_orig[i,1:])
        list2 = list(indices_red[i,1:])
        intersection = len(list(set(list1).intersection(list2)))
        union = (len(list1) + len(list2)) - intersection
        jaccards.append(float(intersection) / union)
    
    return (jaccards)


def get_distortion(X_orig, X_red, k, precomputed=[False, False]):
    '''
    Computes Distortion at k for each point in X_orig and X_red
    
    Distortion = ABS ( LOG [ (D_furthest/D_nearest)_orig / (D_furthest/D_nearest)_red ] )
    Distortions normalized by the maximum in entire dataset to be from [0,1] and reframed so 1 is best
    '''
    # INIT NEAREST NEIGHBORS
    if precomputed[0] is False:
        nbrs = NearestNeighbors(n_neighbors=k).fit(X_orig)
    else:
        nbrs = NearestNeighbors(n_neighbors=k, metric='precomputed').fit(X_orig)
    distances_orig, indices_orig = nbrs.kneighbors(X_orig)
    
    if precomputed[0] is False:
        nbrs = NearestNeighbors(n_neighbors=k).fit(X_red)
    else:
        nbrs = NearestNeighbors(n_neighbors=k, metric='precomputed').fit(X_red)
    distances_red, indices_red = nbrs.kneighbors(X_red)
    
    # COMPUTE DISTORTION
    distortions = []
    for i in range(X_orig.shape[0]):
        orig_ratio = np.max(distances_orig[i,1:]) / np.min(distances_orig[i,1:])
        red_ratio = np.max(distances_red[i,1:]) / np.min(distances_red[i,1:])
        distortions.append(np.abs(np.log(orig_ratio/red_ratio)))
        
    distortions = np.array(distortions)/np.max(distortions)
    distortions = 1-distortions
    
    return (distortions)


def get_mean_projection_error(X_orig, X_red):
    '''
    Computes mean projection error (MPE) modified from aggregated projection error by Martins et al., 2014
    
    MPE_i = MEAN_j { ABS[ D[i,j]_orig / max(D[i,j]_orig) - D[i,j]_red / max(D[i,j]_red) ] }
    
    Normalized to [0,1] and then reframed so 1 is best
    '''
    orig_distance = pairwise_distances(X_orig)
    red_distance = pairwise_distances(X_red)
    
    # normalize distances
    for i in range(orig_distance.shape[0]):
        orig_distance[i,:] = orig_distance[i,:]/np.max(orig_distance[i,:])
        red_distance[i,:] = red_distance[i,:]/np.max(red_distance[i,:])
    
    # compute projection errors and then MPE
    projection_errors = np.abs(orig_distance-red_distance)
    MPEs = np.mean(projection_errors, axis=1)
    MPEs = 1-MPEs/np.max(MPEs)
    return(MPEs)


def get_stretch(X_orig, X_red):
    '''
    Implemented according to Aupetit, 2007:
    
    stretch_i = [ u_i - min_k {u_k} ] / [ max_k {u_k} - min_k {u_k} ] 
    
    u_i = SUM_j D_ij^+  , D_ij^+ = max {-(D_orig_ij - D_red_ij), 0 }  -- D here being Euclidean distance matrix
    
    stretchs reframed so 1 is best
    '''
    orig_distance = pairwise_distances(X_orig)
    red_distance = pairwise_distances(X_red)
    
    # normalize distances
    for i in range(orig_distance.shape[0]):
        orig_distance[i,:] = orig_distance[i,:]/np.linalg.norm(orig_distance[i,:])
        red_distance[i,:] = red_distance[i,:]/np.linalg.norm(red_distance[i,:])
    
    D_neg = orig_distance-red_distance
    D_neg[D_neg>0] = 0
    D_neg = -D_neg
    
    U = np.sum(D_neg, axis=1)
    stretchs = (U-np.min(U))/(np.max(U)-np.min(U))
    stretchs = 1-stretchs
    
    return(stretchs)


def concordance(df, X_orig, method, k=None, bootstrap_number=-1):
    '''
    Computes concordance scores between the projections in df and X_orig
    
    Arguments:
        df = pandas dataframe: output of boot.generate()
        X_orig = nxp numpy array that is the original data from which df was generated
        method = str: 'spearman', 'jaccard', 'distortion',
                 'mean_projection_error', 'stretch'
        k = int, neighborhood size to consider (jaccard, distortion, projection_precision_score, precision, recall)
        bootstrap_number = int, index of bootstrap to compute metrics for; defaults to -1 which is the original/unbootstrapped projection
        
    Returns:
        metrics = numpy array with quality score for each row of df (according to the method specified) [0 is bad, 1 is good]
    '''
    # retrieve embeddings
    X_red = df[df["bootstrap_number"]==bootstrap_number][["x1","x2"]].values
    
    # shuffle X_orig to matching format
    boot_idxs = df[df["bootstrap_number"]==bootstrap_number]["original_index"].values
    X_orig = X_orig[boot_idxs,:]
    
    assert X_orig.shape[0] == X_red.shape[0], "Error: number of observations are not consistent"
    
    # set k to a globally relevant value if None
    if k is None:
        k = round(X_orig.shape[0]/2-1)
    if k < 5:
        raise Exception('k needs to be >= 5 or number of observations in X is too small')
    
    if method == 'spearman':
        orig_distance = pairwise_distances(X_orig)
        red_distance = pairwise_distances(X_red)
        metrics = [spearmanr(orig_distance[i,:],red_distance[i,:])[0] for i in range(red_distance.shape[0])]
    elif method == 'jaccard':
        metrics = get_jaccard(X_orig, X_red, k)
    elif method == 'distortion':
        metrics = get_distortion(X_orig, X_red, k)
    elif method == 'mean_projection_error':
        metrics = get_mean_projection_error(X_orig, X_red)
    elif method == 'stretch':
        metrics = get_stretch(X_orig, X_red)
    else:
        raise Exception("method not recognized")
        
    return(metrics)


def ensemble_concordance(df, X_orig, methods=None, k=None, bootstrap_number=-1, verbose=True):
    '''
    Compute emsemble concordation via spectral meta-weights
    
    Arguments:
        df = pandas dataframe: output of boot.generate()
        X_orig = nxp numpy array that is the original data from which df was generated
        methods = list of strings specifying the metrics to use
            Defaults to all the available metrics (except for precision and recall)
        k = int, neighborhood size to consider (jaccard, distortion, projection_precision_score, precision, recall)
        bootstrap_number = int, index of bootstrap to compute metrics for; defaults to -1 which is the original/unbootstrapped projection
        verbose = True or False, whether to return warnings about negative spectral weights
    
    Returns:
        ensemble_metric = numpy array of ensemble concordance scores
        pointwise_metrics_list = list of concordance score arrays corresponding to pointwise_metrics_labels
    '''
    if methods is None:
        methods = ['spearman', 'jaccard', 'distortion', 'mean_projection_error', 'stretch']
    
    # compute individual metrics
    pointwise_metrics_list = []

    for metric in tqdm(methods):
        m = concordance(df, X_orig, method=metric, k=k, bootstrap_number=bootstrap_number)
        pointwise_metrics_list.append(np.array(m))
    
    # compute correlation matrix and eigendecomposition
    mat = np.corrcoef(pointwise_metrics_list)
    w, v = np.linalg.eig(mat)
    pc1_score = v[:,0]
    
    # check pc1 for one sign only
    summed_signs = np.abs(np.sum(np.sign(pc1_score)))
    if verbose is True:
        if summed_signs == len(pc1_score):
            print ("PC1 is same signed")
        else:
            print ("Warning: PC1 is mixed signed")
            
    # compute meta-uncertainty
    weighted_metrics_list = [pointwise_metrics_list[i]*np.abs(pc1_score)[i] for i in range(len(pointwise_metrics_list))]
    ensemble_metric = np.sum(weighted_metrics_list,axis=0)/np.sum(np.abs(pc1_score))
    
    return(ensemble_metric, pointwise_metrics_list)


"""
Bootstrap visualization module
- Tools for generating bootstrap dimensionality reduction (DR) visualizations
- Standalone methods for DR visualizations
"""

# author: Eric David Sun <edsun@stanford.edu>
# (C) 2022 
from __future__ import print_function

import sklearn.manifold as skm
from sklearn.decomposition import PCA, TruncatedSVD
import numpy as np
import pandas as pd
import os
import umap
from scipy.spatial.transform import Rotation
from tqdm import tqdm
from joblib import Parallel, delayed
import warnings
warnings.filterwarnings("ignore")
import time


def dimensionality_reduction(X, method, **kwargs):
    '''
    Runs a given DR method (specified by method) for the data X (n by p matrix/array)
    
    Uses default arguments unless additional argument values specified in **kwargs
    '''
    # initialize 2D DR method
    if isinstance(method, str):
        if method == "tsne":
            X_embedded = skm.TSNE(n_components=2, verbose=0, **kwargs).fit_transform(X)
        elif method == "mds":
            X_embedded = skm.MDS(n_components=2, verbose=0, **kwargs).fit_transform(X)
        elif method == "lle":
            X_embedded = skm.LocallyLinearEmbedding(n_components=2, **kwargs).fit_transform(X)
        elif method == "mlle":
            X_embedded = skm.LocallyLinearEmbedding(n_components=2, method='modified', **kwargs).fit_transform(X)
        elif method == "isomap":
            X_embedded = skm.Isomap(n_components=2, **kwargs).fit_transform(X)
        elif method == "umap":
            X_embedded = umap.UMAP(n_components=2, verbose=0, **kwargs).fit_transform(X)
        elif method == "pca":
            X_embedded = PCA(n_components=2, **kwargs).fit_transform(X)
        else:
            raise Exception(method+" is not a recognized dimensionality reduction algorithm")
            
    # apply DR and obtain 2D embedding
    else:
        X_embedded = method.fit_transform(X)
    
    return(X_embedded)


def bootstrap (X, method, B, sigma_noise=None, no_bootstrap=False, random_seed=None, num_jobs=None, use_n_pcs=False, subsample=False, **kwargs):
    '''
    Creates n bootstrap data from X and creates a DR visualizastion for each of them.
    
    Arguments:
        See generate() for details
    
    Returns:
        X_embedded_list = list of the 2D DR visualization embeddings (numpy arrays)
        bootstrap_indices_list = list of numpy arrays indicating the bootstrap row indices
    '''

    X_embedded_list = []
    bootstrap_indices_list = []
    
    # generate sequence of random states if random_seed is specified
    if isinstance(random_seed, int):
        seeded_rand1 = np.random.RandomState(random_seed)
        random_seeded_sequence = seeded_rand1.randint(0,1e6,B)
    else:
        random_seeded_sequence = False
    
    # bootstrap DR
    if isinstance(num_jobs, int): # in parallel
        result = Parallel(n_jobs=num_jobs)(delayed(run_one_bootstrap)(X, method, sigma_noise, no_bootstrap,
                                           random_seeded_sequence, b, use_n_pcs, subsample, **kwargs) for b in tqdm(range(B)))
        X_embedded_list = [x[0] for x in result]
        bootstrap_indices_list = [x[1] for x in result]
    else: # using only one core
        for b in tqdm(range(B)):
            X_embedded, boot_idxs = run_one_bootstrap(X, method, sigma_noise, no_bootstrap,
                                           random_seeded_sequence, b, use_n_pcs, subsample, **kwargs)
            X_embedded_list.append(X_embedded)
            bootstrap_indices_list.append(boot_idxs)
    
    return(X_embedded_list, bootstrap_indices_list)


def run_one_bootstrap(X, method, sigma_noise=None, no_bootstrap=False, random_seeded_sequence=False, b=0, use_n_pcs=False, subsample=False, **kwargs):
    '''
    Method for generating one bootstrap X and one DR visualization of the bootstrap
    
    Arguments:
        random_seeded_sequence = array or list of random seeds to use in generating the bootstrap sample
        b = integer specifying the index of random see in random_seeded_sequence to use for generating the bootstrap
        See generate() for more details
    
    Returns:
        X_embedded = 2D DR visualization embedding (numpy array)
        boot_idxs = numpy array indicating the bootstrap row indices
    '''
    # Create bootstrap X
    if subsample is False:
        if no_bootstrap is True: # don't bootstrap (will use intrinsic stochasticity of DR algorithm (if any) only)
            boot_X = X.copy()
            boot_idxs = np.arange(X.shape[0]) # set indices to be the original indices
        elif random_seeded_sequence is not False: # use specified random_seeded_sequence to generate bootstrap X
            seeded_rand2 = np.random.RandomState(random_seeded_sequence[b])
            boot_idxs = seeded_rand2.randint(0,X.shape[0],X.shape[0])
            boot_X = X.copy()[boot_idxs,:] 
        else: # if no random_seeded_sequence, use default random process
            boot_idxs = np.random.randint(0,X.shape[0],X.shape[0])
            boot_X = X.copy()[boot_idxs,:]
    # Subsample instead of bootstrapping
    else:
        if random_seeded_sequence is not False: # use specified random_seeded_sequence to generate subsample of X
            seeded_rand2 = np.random.RandomState(random_seeded_sequence[b])
            boot_idxs = seeded_rand2.choice(X.shape[0], subsample, replace=False)
            boot_X = X.copy()[boot_idxs,:] 
        else: # if no random_seeded_sequence, use default random process
            boot_idxs = np.random.choice(X.shape[0], subsample, replace=False)
            boot_X = X.copy()[boot_idxs,:]
        
    # add Gaussian noise to alleviate duplicate issues if specified
    if sigma_noise is not None:
        boot_X += np.random.normal(0,sigma_noise,boot_X.shape)
        
    # run TruncatedSVD if specified to do a first pass DR with PCA and take use_n_pcs top principal components for DR visualization
    if use_n_pcs is not False:
        boot_X = TruncatedSVD(n_components=use_n_pcs).fit_transform(boot_X)
    
    # generate DR visualization embedding
    X_embedded = dimensionality_reduction(boot_X, method, **kwargs)


    return(X_embedded, boot_idxs)


def generate(X, method, Y=None, B=0, sigma_noise=None, no_bootstrap=False, random_seed=None, save=False,
             num_jobs=None, use_n_pcs=False, subsample=False, return_times=False, **kwargs):
    '''
    Main method for generating aligned bootstrap visualizations, which are the input elements for dynamic visualization.
    
    Arguments:
        X = (n x p) numpy array where rows are observations, columns are features
        method = string, dimensionality reduction method to use; options include:
            "tsne", "mds", "lle", "mlle", "isomap", "umap", "pca"
        Y = pandas dataframe with same number of rows as X and columns containing relevant metadata to propagate to output
        B = integer, number of bootstraps to generate; if B==0: generates only for the original X
        sigma_noise = None or float, if float, adds zero-centered Gaussian noise to each bootstrap sample with standard deviation sigma_noise
        no_bootstrap = True or False, whether to bootstrap sample for each iteration or not
        random_seed = None or int, if int, uses that value to generate random sequence (i.e. bootstrap sequence will be the same)
        save = False or str, if str, path to save resulting Pandas dataframe as CSV
        num_jobs = None, -1, or >=1 int; if not None, runs multiprocessing with n_jobs, if n_jobs=-1, then uses all available
        use_n_pcs = False or int, specifying to apply PCA and keep to use_n_pcs components to use for method
        subsample = False or int, specifying whether to subsample INSTEAD OF bootstrapping with integer corresponding to size of subsample to take
        return_times = True or False; if not False, returns a dictionary of run times broken down by components as the second output
    
    Returns:
        output = Pandas dataframe with "x1", "x2", "bootstrap_number", "original_index" as columns, along with columns of Y
                    2D embedding is (x1, x2)
    '''
    # process results into dataframe
    output = pd.DataFrame() # init df to be merged onto
    
    # DR on original dataset
    original_embedding = dimensionality_reduction(X, method, **kwargs) # Ex: can specify random_state in **kwargs to remove stochasticity
    
    # reference points is the original dataset
    points0 = np.hstack((original_embedding, np.zeros(original_embedding.shape[0]).reshape(original_embedding.shape[0],1)))# append uniform 3rd dimension

    # add basic info
    output["x1"] = points0[:,0]-np.mean(points0[:,0]) # center reference visualization at (0,0)
    output["x2"] = points0[:,1]-np.mean(points0[:,1])
    output["original_index"] = np.arange(len(points0[:,0]))
    output["bootstrap_number"] = -1

    # add metadata
    if isinstance(Y, pd.DataFrame):
        for col in Y.columns:
            output[col] = Y[col].values
    
    # keep track of run times
    rt_dict = {}
    
    # bootstrap
    start_time = time.time()
    if B > 0:
        bootstrap_embedding_list, bootstrap_indices_list = bootstrap(X, method, B, sigma_noise, no_bootstrap, 
                                                                    random_seed, num_jobs, use_n_pcs, subsample, **kwargs)
    bootstrap_time = time.time() - start_time
    rt_dict["bootstrapped_DR"] = bootstrap_time
    
    # add bootstraps
    start_time = time.time()
    for i in range(len(bootstrap_embedding_list)):
        
        new_df = pd.DataFrame() # new df to merge onto original df
        
        points = bootstrap_embedding_list[i]
        points = np.hstack((points, np.zeros(points.shape[0]).reshape(points.shape[0],1)))# append uniform 3rd dimension
        boot_idxs = bootstrap_indices_list[i]
        
        # rigid alignment w/ 3d rotation (Kabsch)
        ref_points = points0[boot_idxs,:]
        points[:,0] = points[:,0]-np.mean(points[:,0])
        points[:,1] = points[:,1]-np.mean(points[:,1])
        r = Rotation.align_vectors(ref_points, points)[0]
        rpoints = r.apply(points)
        
        # add basic info
        new_df["x1"] = rpoints[:,0]
        new_df["x2"] = rpoints[:,1]
        new_df["original_index"] = boot_idxs
        new_df["bootstrap_number"] = i
        
        # add metadata
        if isinstance(Y, pd.DataFrame):
            for col in Y.columns:
                new_df[col] = Y[col].values[boot_idxs]
        
        # merge to original dataframe
        output = pd.concat([output, new_df], axis=0)
    
    align_time = time.time() - start_time
    rt_dict["alignment_DR"] = align_time
    
    # save output
    if save is not False:
        output.to_csv(save, index=False)
    
    if return_times is False:
        return(output)
    else:
        return(output, rt_dict)
        
        
        


"""
Additional utilities module
- Tools for converting to/from AnnData format
"""

# author: Eric David Sun <edsun@stanford.edu>
# (C) 2022 
from __future__ import print_function

import numpy as np
import pandas as pd
import os
import warnings
warnings.filterwarnings("ignore")

def convert_anndata (adata, obs, obsm="X_pca", n_components=None):
    '''
    Converts an AnnData object into X and Y inputs for boot.generate()
    
    Arguments:
        adata = AnnData object with all metadata of interest in anndata.obs and has adata.X and adata.obsm["X_pca"]
        obs = list of str, str correspond to a key in adata.obs
        obsm = str, str correspond to a key in adata.obsm (i.e. "X_pca") -- if no matches found, will default to adata.X
        n_components = None or int, if int and obsm="X_pca" is found, then take first n_components of X_pca
        
    Returns:
        X = numpy array with the same number of rows as adata.X and columns depending on obsm
        Y = Pandas dataframe with columns specified by obs, same number of rows as X
    '''
    # get X from adata.obsm or adata.X
    obsm_keys = list(adata.obsm.keys())
    if obsm in obsm_keys:
        X = adata.obsm[obsm]
        if obsm == "X_pca":
            X = X[:,:n_components]
    else:
        X = adata.X
    
    # get Y from adata.obs
    Y = adata.obs[obs]
    
    return (X, Y)

def regenerate_anndata (adata, df, bootstrap_number, obsm):
    '''
    Converts AnnData to match df at specified bootstrap_number
    
    Will replace obsm (e.g. "X_umap") with df[["x1", "x2"]]
    Will reorganize all other metadata and X to match bootstrap indices
    
    Arguments:
        adata = original AnnData
        df = pandas dataframe, output of boot.generate()
        bootstrap_number = int, index to use to regenerate anndata (i.e. an integer appearing in df["bootstrap_number"])
        obsm = str, key to replace adata.obsm[key] with df[["x1", "x2"]]
    
    Returns:
        adata_copy = new AnnData 
    '''
    bootstrap_idxs = df[df["bootstrap_number"] == bootstrap_number]["original_index"].values
    adata_copy = adata.copy()[bootstrap_idxs]
    adata_copy.obsm[obsm] = df[df["bootstrap_number"] == bootstrap_number][["x1", "x2"]].values # update DR embedding

    return(adata_copy)


# rudimentary test to see if bootstrapping and interactive visualization is functioning
# similar to tutorial.ipynb


# import packages
from dynamicviz import boot, viz, score
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
from scipy.stats import spearmanr
from sklearn.datasets import make_s_curve


if not os.path.exists('tests/outputs'):
    os.makedirs('tests/outputs')

# load data
S_X, S_y = make_s_curve(1000, random_state=0)
S_y = pd.DataFrame(S_y, columns=["label"])


# with parallelization
try:
    out = boot.generate(S_X, Y=S_y, method="tsne", B=4, save=False, random_seed=452, random_state=452, num_jobs=2)
except:
    print("Warning: Error encountered parallelization of boot.generate() -- ignore if only using one core")

# without parallelization
try:
    out = boot.generate(S_X, Y=S_y, method="tsne", B=4, save=False, random_seed=452, random_state=452)
except:
    raise Exception("Error encountered in boot.generate")


# check output
truth = pd.read_csv("tests/outputs/truth.csv")
assert round(out["x1"].values[333],4) == round(truth["x1"].values[333],4), "Output dataframe does not match truth.csv"
assert round(out["x2"].values[1111],4) == round(truth["x2"].values[1111],4), "Output dataframe does not match truth.csv"

# make interactive visualization
try:
    fig = viz.interactive(out, 'label', show=False, save=False, alpha=0.5, legend_title="Cell type", dpi=150)
except:
    print ("Warning: Error encountered for viz.interactive()")

# make animated visualization
try:
    fig = viz.animated(out, 'label', save=False)
except:
    print ("Warning: Error encountered for viz.animated()")

# make stacked visualization
try:
    fig = viz.stacked(out, 'label', show=False, save=False,
                    xlabel="t-SNE 1", ylabel="t-SNE 2", dpi=150, marker="x", s=20, show_legend=True, solid_legend=True, cmap='hot')
except:
    print ("Warning: Error encountered for viz.stacked()")
                    
# global variance score
try:
    variance_scores = score.variance(out, method="global")
except:
    print ("Warning: Error encountered for score.variance(method='global')")
    
# check variance scores
truth = np.genfromtxt("tests/outputs/truth_variances.txt")
assert round(variance_scores[333],4) == round(truth[333],4), "Variance scores do not match truth_variances.txt"
    
# random variance score
try:
    variance_scores = score.variance(out, method="random", k=50)
except:
    print ("Warning: Error encountered for score.variance(method='random')")
    
# concordance score test
for method in ["spearman", "distortion", "jaccard", "mean_projection_error", "stretch"]:
    try:
        concordance_scores = score.concordance(out, S_X, method=method, k=50, bootstrap_number=-1)
    except:
        print ("Warning: Error encountered for score.concordance(method='"+method+"')")


